-- Speedway v0.2
-- By Kimina of Drenden-US
--
-- Tracks laps around Dalaran (and maybe other cities!)
-- 
--    Copyright (c) 2009 Kimina of Drenden-US
--    This file is part of Speedway.
--
--    Speedway is free software: you can redistribute it and/or modify
--    it under the terms of the GNU General Public License as published by
--    the Free Software Foundation, either version 3 of the License, or
--    (at your option) any later version.
--
--    Speedway is distributed in the hope that it will be useful,
--    but WITHOUT ANY WARRANTY; without even the implied warranty of
--    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--    GNU General Public License for more details.
--
--    You should have received a copy of the GNU General Public License
--    along with Speedway.  If not, see <http://www.gnu.org/licenses/>.
-- 

local addon = {};		-- The addon
local cities = {};		-- Listing of cities and their gates
addon.cities = cities;

local gateMT = {__index = {}};	-- Metatable used for gates
local registerEvent, unregisterEvent;
local recordCrossing, addRotation;

local UPDATE_RATE = 0.5;	-- Update rate of OnUpdate

local Astrolabe = DongleStub("Astrolabe-0.4");

Speedway_Laps = {};

--[[========================================
--|| INITIALIZATION
--||======================================]]
local loaded = false;
function addon.ADDON_LOADED()
  if not loaded then
	-- Fill in the list of cities, points are clockwise
	cities[{ZoneText = "Dalaran", key = "Dalaran"}] = setmetatable({
		{.4615, .6173, .4850, .5686}, -- Near Eventhide
		{.3961, .4298, .4331, .4445}, -- Violet Citadel
		{.5090, .3815, .5407, .3466}, -- Horde area
		{.5674, .4856, .5955, .5011}, -- Violet Hold
	}, gateMT);

	loaded = true;
  end
end


--[[========================================
--|| CORE
--||======================================]]
-- Fired on every frame; delay to only every UPDATE_RATE secs
local lastTime, lastX, lastY = 0;
function addon.OnUpdate(frame, dt)
	-- Check for only every few seconds
	lastTime = lastTime + dt;
	if (lastTime < UPDATE_RATE) then
		return;
	end

	lastTime = 0;

	-- Update X/Y
	local _, _, x, y = Astrolabe:GetCurrentPlayerPosition();

	if not (lastX and lastY) then
		lastX, lastY = x, y;
		return;
	end

	-- Compare coords against last, see if we passed a gate
	local zone = GetZoneText();
	local subzone = GetSubZoneText();

	-- Go through all our places to see if we have a match
	for k,v in pairs(cities) do
		if (not k.ZoneText or k.ZoneText == zone) and
	           (not k.SubzoneText or k.SubzoneText == subzone) then

		   -- Match! Check for gate crossing
		   recordCrossing(k, v:GetGateCrossing(lastX, lastY, x, y));
	        end
	end
	
	lastX, lastY = x, y;	
end

-- Records a crossing across a gate
local rotation, lastZone, lastGate = 0;
recordCrossing = function(zone, n)
	if not n then return; end	-- There was no crossing

	if lastZone ~= zone then	-- Mismatch in zones
		rotation = 0;
		lastZone = zone;
		lastGate = n;
		return;
	end

	if n == 1 and lastGate == #cities[zone] or n == lastGate+1 then
		rotation = rotation + 1 / #cities[zone];
		-- Clockwise rotation
		if rotation >= 1 then
			addRotation(zone, 1);
			rotation = 0;
		end
	elseif lastGate == 1 and n == #cities[zone] or n == lastGate-1 then
		rotation = rotation - 1 / #cities[zone];
		-- Counter-clockwise rotation
		if rotation <= -1 then
			addRotation(zone, -1);
			rotation = 0;
		end
	else
		-- Invalid rotation
		rotation = 0;
	end

	lastGate = n;
end

-- Adds to the rotation totals
-- direction = 1 if clockwise, -1 if counter-clockwise
addRotation = function(zone, direction)
	local sv = Speedway_Laps[zone.key] or {cw = 0, ccw = 0};

	if (direction == 1) then
		sv.cw = sv.cw + 1;
	elseif (direction == -1) then
		sv.ccw = sv.ccw + 1;
	else
		error("Invalid direction");
	end

	-- In case this is new
	Speedway_Laps[zone.key] = sv;
end

--[[========================================
--|| GATE MAGIC
--||======================================]]
-- Determines if this is a gate crossing and returns the 
-- gate ID or nil
function gateMT.__index:GetGateCrossing(x1, y1, x2, y2)
	-- Process by constructing 2 line segments and
	-- determining if they intersect
	
	-- ua = (x4 - x3)(y1 - y3) - (y4 - y3)(x1 - x3)
	--      ---------------------------------------
	--      (y4 - y3)(x2 - x1) - (x4 - x3)(y2 - y1)
	-- 
	-- ub = (x2 - x1)(y1 - y3) - (y2 - y1)(x1 - x3)
	--      ---------------------------------------
	--      (y4 - y3)(x2 - x1) - (x4 - x3)(y2 - y1)
	--
	-- x = x1 + (ua or ub)(x2 - x1)
	-- y = y1 + (ua or ub)(y2 - y1)
	--
	-- The X and Y coord of the intersection is not really
	-- necessary for our purposes. All we care about is whether
	-- or not it intersects. Either ua or ub (or both) must be
	-- between 0 and 1, inclusively. (Be prepared to handle the
	-- cases where denominator is 0 for parallel, or 0/0 for coincident)
	
	for i,v in ipairs(self) do 
		assert(#v == 4);
		local x3, y3, x4, y4 = unpack(v);

		local denom = (y4 - y3) * (x2 - x1) - (x4 - x3) * (y2 - y1);
		local ua = ((x4 - x3) * (y1 - y3) - (y4 - y3) * (x1 - x3))/denom;
		local ub = ((x2 - x1) * (y1 - y3) - (y2 - y1) * (x1 - x3))/denom;
--		if (ua >= 0 and ua <= 1) then print(ua) end
--		if (ub >= 0 and ub <= 1) then print(ub) end

--		print(x1 .. ", " .. y1 .. " => " .. x2 .. ", " .. y2);
--		print(x3 .. ", " .. y3 .. " => " .. x4 .. ", " .. y4);

		if (ua >= 0 and ua <= 1 or ua ~= ua) and (ub >= 0 and ub <= 1 or ub ~= ub) then
			return i;
		end
	end

	return nil;	-- Not really necessary, but we do it anyway	
end

--[[=========================================
--|| DISPLAY OF LAPS
--||=======================================]]
SlashCmdList["SPEEDWAY"] = function(data)
  data = data:match("^%s*(.*)%s*$") or "";
  if (data:lower() == "reset") then
	Speedway_Laps = {};
	print("|cffaaaaffSpeedway Laps Reset|r");
  elseif (data == "") then
	local printed = false;
	for k,v in pairs(Speedway_Laps) do
		printed = true;
		print("|cffaaaaff" .. k .. ": " .. "|r|cffffffff" ..
			v.cw .. "|r|cffaaaaff Clockwise, |r|cffffffff" ..
			v.ccw .. "|r|cffaaaaff Counter-Clockwise Laps");
	end

	if not printed then
		print("|cffaaaaffYou haven't made any laps! Go unbore yourself!");
	end
  else
        -- Find this one
	local lapData;
	for k,v in pairs(Speedway_Laps) do
		if k:lower() == data:lower() then
			lapData = v;
		end
	end

	if not lapData then
		print("|cffaaaaff" .. data .. "|r |cffffaaaahas no lap data|r");
	else
		print("|cffaaaaff" .. data .. ": " .. "|r|cffffffff" ..
			lapData.cw .. "|r|cffaaaaff Clockwise, |r|cffffffff" ..
			lapData.ccw .. "|r|cffaaaaff Counter-Clockwise Laps");
	end
  end
end

SLASH_SPEEDWAY1 = "/speedway";
	


--[[========================================
--|| EVENT HANDLING
--||======================================]]
local eventFrame = CreateFrame("FRAME");
-- Event dispatcher
eventFrame:SetScript("OnEvent", function(self, evt, ...)
	if addon[evt] then
		return addon[evt](...);
	end
end);

eventFrame:SetScript("OnUpdate", addon.OnUpdate);

-- Registers an event for the frame
registerEvent = function(evt)
	if not eventFrame:IsEventRegistered(evt) then
		eventFrame:RegisterEvent(evt);
	end
end

-- Unregisters an event from the frame
unregisterEvent = function(evt)
	if eventFrame:IsEventRegistered(evt) then
		eventFrame:UnregisterEvent(evt);
	end
end

registerEvent("ADDON_LOADED");
